# Docs: https://github.com/jonluca/anubis

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://jonlu.ca/anubis/subdomains"


# Query: Domain
# Return: Subdomains
def req_anubis_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching Anubis...", ihunt.verbose)

    url = BASE_URL + f"/{ihunt.query.value}"

    try:
        resp = requests.get(url)
        if resp.status_code == 200:
            with lock:
                subdomains = resp.json()
                for subdomain in subdomains:
                    if ihunt.data.subdomains is None:
                        ihunt.data.subdomains = [subdomain]
                    else:
                        if subdomain not in ihunt.data.subdomains:
                            ihunt.data.subdomains.append(subdomain)
    except Exception as e:
        echo(f"[x] Anubis API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching Anubis...", ihunt.verbose)



# Docs: https://eva.pingutil.com/

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://api.eva.pingutil.com/email"

# Query: Email
# Return: Verification
def req_eva_email(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching EVA...", ihunt.verbose)

    url = BASE_URL
    params = {
        "email": ihunt.query.value,
    }

    try:
        resp = requests.get(url, params=params)

        if resp.status_code == 200:
            with lock:
                d = resp.json()
                if d["success"] != "success":
                    data = d["data"]
                    if ihunt.data.email is None:
                        ihunt.data.email = data["email_address"]
                    if ihunt.data.domain is None:
                        ihunt.data.domain = data["domain"]
                    if ihunt.data.gibberish is None:
                        ihunt.data.gibberish = data["gibberish"]
                    if ihunt.data.disposable is None:
                        ihunt.data.disposable = data["disposable"]
                    if ihunt.data.webmail is None:
                        ihunt.data.webmail = data["webmail"]
                    if ihunt.data.spam is None:
                        ihunt.data.spam = data["spam"]
                    return
    except Exception as e:
        echo(f"[x] EVA API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching EVA.", ihunt.verbose)


# # Docs: https://developer.shodan.io/api

# from dataclasses import dataclass
# import requests

# BASE_URL = "https://api.shodan.io"


# # Query: Domain
# # Return: Info
# def req_shodan_domain(domain: str, apikey: str) -> None:
#     url = BASE_URL + f"/dns/domain/{domain}?key={apikey}"
#     resp = requests.get(url)
#     if resp.status_code == 200:
#         data = resp.json()
#         d_data : list[dict[str, str]]  = data["data"]
#         subdomains: list[Subdomain] = []
#         for d in d_data:
#             subdomains.append(
#                 Subdomain(
#                     domain=d["subdomain"],
#                     type=d["type"],
#                     value=d["value"],
#                     last_seen=d["last_seen"],
#                 )
#             )
#         # return ShodanDomainResponse(
#         #     domain=data["domain"],
#         #     subdomains=subdomains,
#         # )
#         return None
#     else:
#         return None


# # Query: IP
# # Return: Info
# def req_shodan_ip(ip: str, apikey: str) -> None:
#     url = BASE_URL + f"/host/{ip}?key={apikey}"
#     resp = requests.get(url)
#     data = resp.json()
#     if resp.status_code == 200:
#         # return ShodanIpResponse(
#         #     region_code=data["region_code"],
#         #     ip=data["ip"],
#         #     postal_code=data["postal_code"],
#         #     country_code=data["country_code"],
#         #     city=data["city"],
#         #     dma_code=data["dma_code"],
#         #     last_update=data["last_update"],
#         #     latitude=data["latitude"],
#         #     tags=data["tags"],
#         #     area_code=data["are_code"],
#         #     country_name=data["country_name"],
#         #     hostnames=data["hostnames"],
#         #     org=data["org"],
#         #     asn=data["asn"],
#         #     isp=data["isp"],
#         #     longitude=data["longitude"],
#         #     country_code3=data["country_code3"],
#         #     domains=data["domains"],
#         #     ip_str=data["ip_str"],
#         #     os=data["os"],
#         #     ports=data["ports"],
#         # )
#         return None
#     else:
#         return None




# Docs: https://hunter.io/api-documentation/v2

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://api.hunter.io/v2"


# Query: Domain
# Return: Info
def req_hunter_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching Hunter...", ihunt.verbose)

    url = BASE_URL + f"/domain-search"
    params = {
        "domain": ihunt.query.value,
        "api_key": ihunt.apikeys.hunter,
    }

    try:
        resp = requests.get(url, params=params)
        if resp.status_code == 200:
            with lock:
                d = resp.json()["data"]
                if ihunt.data.domain is None:
                    ihunt.data.domain = d["domain"]
                if ihunt.data.registrant_organization is None:
                    ihunt.data.registrant_organization = d["organization"]
                if ihunt.data.registrant_desc is None:
                    ihunt.data.registrant_desc = d["description"]
                if ihunt.data.registrant_industry is None:
                    ihunt.data.registrant_industry = d["industry"]
                if ihunt.data.registrant_twitter is None:
                    ihunt.data.registrant_twitter = d["twitter"]
                if ihunt.data.registrant_facebook is None:
                    ihunt.data.registrant_facebook = d["facebook"]
                if ihunt.data.registrant_linkedin is None:
                    ihunt.data.registrant_linkedin = d["linkedin"]
                if ihunt.data.registrant_instagram is None:
                    ihunt.data.registrant_instagram = d["instagram"]
                if ihunt.data.registrant_youtube is None:
                    ihunt.data.registrant_youtube = d["youtube"]
                if ihunt.data.registrant_country_code is None:
                    ihunt.data.registrant_country_code = d["country"]
                if ihunt.data.registrant_state is None:
                    ihunt.data.registrant_state = d["state"]
                if ihunt.data.registrant_city is None:
                    ihunt.data.registrant_city = d["city"]
                if ihunt.data.registrant_street is None:
                    ihunt.data.registrant_street = d["street"]
                if ihunt.data.registrant_postal_code is None:
                    ihunt.data.registrant_postal_code
                if len(d["emails"]) > 0:
                    for email in d["emails"]:
                        if ihunt.data.emails is None:
                            ihunt.data.emails = [email["value"]]
                        else:
                            ihunt.data.emails.append(email["value"])
    except Exception as e:
        echo(f"[x] Hunter API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching Hunter.", ihunt.verbose)


# Query: Email
# Return: Verification
def req_hunter_email(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching Hunter...", ihunt.verbose)

    url = BASE_URL + f"/email-verifier"
    params = {
        "email": ihunt.query.value,
        "api_key": ihunt.apikeys.hunter,
    }

    try:
        resp = requests.get(url, params=params)

        if resp.status_code == 200:
            with lock:
                d = resp.json()["data"]
                if ihunt.data.gibberish is None:
                    ihunt.data.gibberish = d["gibberish"]
                if ihunt.data.disposable is None:
                    ihunt.data.disposable = d["disposable"]
                if ihunt.data.webmail is None:
                    ihunt.data.webmail = d["webmail"]
                if ihunt.data.mx_records is None:
                    ihunt.data.mx_records = d["mx_records"]
                if ihunt.data.smtp_server is None:
                    ihunt.data.smtp_server = d["smtp_server"]
                if ihunt.data.smtp_check is None:
                    ihunt.data.smtp_check = d["smtp_check"]
                if ihunt.data.accept_all is None:
                    ihunt.data.accept_all = d["accept_all"]
                if ihunt.data.block is None:
                    ihunt.data.block = d["block"]
    except Exception as e:
        echo("[x] Hunter API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching Hunter.", ihunt.verbose)


# # Docs: https://urlscan.io/docs/api/

import requests

# BASE_URL = "https://urlscan.io/api/v1/search"


# # Query: Domain
# # Return: Info
# def req_urlscan_domain(domain: str, apikey: str) -> None:
#     url = BASE_URL + f"/?q=domain:{domain}&size=100"
#     headers = {
#         'API-KEY': apikey,
#     }
#     resp = requests.get(url, headers=headers)

#     if resp.status_code == 200:
#         data = resp.json()
#         results = data["results"]
#         if len(results) == 0:
#             return None
        
#         for result in results:
#             task = result["task"]
#             page = result["page"]
#             print(f"task: {task}")
#             print(f"page: {page}")
#             break

#         return None
#     else:
#         return None



# Docs: https://urldna.io/api

import json
import requests
from threading import Lock
import time
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://api.urldna.io/scan"


# Query: URL
# Return: Info
def req_urldna_url(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching urlDNA...", ihunt.verbose)

    url = BASE_URL
    headers = {
        "Authorization": ihunt.apikeys.urldna,
        "Content-Type": "application/json",
    }
    data = {
        "submitted_url": ihunt.query.value,
    }

    try:
        resp = requests.post(url, headers=headers, data=json.dumps(data))
        if resp.status_code != 200:
            return
        with lock:
            d = resp.json()
            scan_id = d["id"]
            url = BASE_URL + f"/{scan_id}"

            # Get scan result (wait for the scan status is "DONE")
            while True:
                echo("[*] Waiting for the scan result...", ihunt.verbose)
                time.sleep(5)

                resp = requests.get(url, headers=headers)
                if resp.status_code != 200:
                    return
            
                d = resp.json()
                scan_status = d["scan"]["status"]

                if scan_status == "RUNNING" or scan_status == "PENDING":
                    continue
                elif scan_status == "ERROR":
                    echo(f"[x] urlDNA API error: The scan failed.", ihunt.verbose)
                    return
                elif scan_status == "DONE":
                    if ihunt.data.cert_issuer is None:
                        ihunt.data.cert_issuer = d["certificate"]["issuer"]
                    if ihunt.data.cert_subject is None:
                        ihunt.data.cert_subject = d["certificate"]["subject"]
                    if ihunt.data.cert_serial_number is None:
                        ihunt.data.cert_serial_number = d["certificate"]["serial_number"]
                    if ihunt.data.console_messages is None:
                        ihunt.data.console_messages = []
                        for msg in d["console_messages"]:
                            ihunt.data.console_messages.append(msg)
                    if ihunt.data.cookies is None:
                        ihunt.data.cookies = d["cookies"]
                    if ihunt.data.dom is None:
                        ihunt.data.dom = d["dom"]
                    if ihunt.data.favicon is None:
                        ihunt.data.favicon = d["favicon"]
                    if ihunt.data.ip is None:
                        ihunt.data.ip = d["ip_address"]["ip"]
                    if ihunt.data.asn is None:
                        ihunt.data.asn = d["ip_address"]["asn"]
                    if ihunt.data.isp is None:
                        ihunt.data.isp = d["ip_address"]["isp"]
                    if ihunt.data.organization is None:
                        ihunt.data.organization = d["ip_address"]["org"]
                    if ihunt.data.country_name is None:
                        ihunt.data.country_name = d["ip_address"]["country"]
                    if ihunt.data.country_code is None:
                        ihunt.data.country_code = d["ip_address"]["country_code"]
                    if ihunt.data.region is None:
                        ihunt.data.region = d["ip_address"]["region"]
                    if ihunt.data.city is None:
                        ihunt.data.city = d["ip_address"]["city"]
                    if ihunt.data.latitude is None:
                        ihunt.data.latitude = str(d["ip_address"]["latitude"])
                    if ihunt.data.longitude is None:
                        ihunt.data.longitude = str(d["ip_address"]["longitude"])
                    if ihunt.data.malicious is None:
                        ihunt.data.malicious = d["malicious"]
                    if ihunt.data.page is None:
                        ihunt.data.page = d["page"]
                    if ihunt.data.screenshot is None:
                        ihunt.data.screenshot = d["screenshot"]
                    if ihunt.data.technologies is None:
                        ihunt.data.technologies = d["technologies"]
                    break
    except Exception as e:
        echo(f"[x] urlDNA API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching urlDNA.", ihunt.verbose)


import subprocess
from threading import Lock
from ..models import Ihunt
from ..stdout import echo


# Query: Domain
# Return: Info
def exec_whois_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Executing whois command...", ihunt.verbose)

    result = subprocess.run(['whois', ihunt.query.value], capture_output=True, text=True)
    lines = result.stdout.splitlines()

    try:
        with lock:
            for line in lines:
                if ':' not in line:
                    continue
                key, val = map(str.strip, line.split(':', 1))
                if key == "Domain Name" and ihunt.data.domain is None:
                    ihunt.data.domain = val
                if key == "Updated Date" and ihunt.data.updated_date is None:
                    ihunt.data.updated_date = val
                if key == "Creation Date" and ihunt.data.creation_date is None:
                    ihunt.data.creation_date = val
                if key == "Registry Expiry Date" and ihunt.data.registry_expiry_date is None:
                    ihunt.data.registry_expiry_date = val
                if key == "Registrar" and ihunt.data.registrar_name is None:
                    ihunt.data.registrar_name = val
                if key == "Registrar IANA ID" and ihunt.data.registrar_iana_id is None:
                    ihunt.data.registrar_iana_id = val
                if key == "Domain Status":
                    if ihunt.data.status is None:
                        ihunt.data.status = [val]
                    else:
                        if val not in ihunt.data.status:
                            ihunt.data.status.append(val)
                if key == "Registrant Name" and ihunt.data.registrant_name is None:
                    ihunt.data.registrant_name = val
                if key == "Registrant Organization" and ihunt.data.registrant_organization is None:
                    ihunt.data.registrant_organization = val
                if key == "Registrant Street" and ihunt.data.registrant_street is None:
                    ihunt.data.registrant_street = val
                if key == "Registrant City" and ihunt.data.registrant_city is None:
                    ihunt.data.registrant_city = val
                if key == "Registrant State/Province" and ihunt.data.registrant_state is None:
                    ihunt.data.registrant_state = val
                if key == "Registrant Postal Code" and ihunt.data.registrant_postal_code is None:
                    ihunt.data.registrant_postal_code = val
                if key == "Registrant Country" and ihunt.data.registrant_country is None:
                    ihunt.data.registrant_country = val
                if key == "Registrant Phone" and ihunt.data.registrant_phone is None:
                    ihunt.data.registrant_phone = val
                if key == "Registrant Fax" and ihunt.data.registrant_fax is None:
                    ihunt.data.registrant_fax = val
                if key == "Registrant Email" and ihunt.data.registrant_email is None:
                    ihunt.data.registrant_email = val
                if key == "Admin Name" and ihunt.data.admin_name is None:
                    ihunt.data.admin_name = val
                if key == "Admin Organization" and ihunt.data.admin_organization is None:
                    ihunt.data.admin_organization = val
                if key == "Admin Street" and ihunt.data.admin_street is None:
                    ihunt.data.admin_street = val
                if key == "Admin City" and ihunt.data.admin_city is None:
                    ihunt.data.admin_city = val
                if key == "Admin State/Province" and ihunt.data.admin_state is None:
                    ihunt.data.admin_state = val
                if key == "Admin Postal Code" and ihunt.data.admin_postal_code is None:
                    ihunt.data.admin_postal_code = val
                if key == "Admin Country" and ihunt.data.admin_country is None:
                    ihunt.data.admin_country = val
                if key == "Admin Phone" and ihunt.data.admin_phone is None:
                    ihunt.data.admin_phone = val
                if key == "Admin Fax" and ihunt.data.admin_fax is None:
                    ihunt.data.admin_fax = val
                if key == "Admin Email" and ihunt.data.admin_email is None:
                    ihunt.data.admin_email = val
                if key == "Tech Name" and ihunt.data.tech_name is None:
                    ihunt.data.tech_name = val
                if key == "Tech Organization" and ihunt.data.tech_organization is None:
                    ihunt.data.tech_organization = val
                if key == "Tech Street" and ihunt.data.tech_street is None:
                    ihunt.data.tech_street = val
                if key == "Tech City" and ihunt.data.tech_city is None:
                    ihunt.data.tech_city = val
                if key == "Tech State/Province" and ihunt.data.tech_state is None:
                    ihunt.data.tech_state = val
                if key == "Tech Postal Code" and ihunt.data.tech_postal_code is None:
                    ihunt.data.tech_postal_code = val
                if key == "Tech Country" and ihunt.data.tech_country is None:
                    ihunt.data.tech_country = val
                if key == "Tech Phone" and ihunt.data.tech_phone is None:
                    ihunt.data.tech_phone = val
                if key == "Tech Fax" and ihunt.data.tech_fax is None:
                    ihunt.data.tech_fax = val
                if key == "Tech Email" and ihunt.data.tech_email is None:
                    ihunt.data.tech_email = val
                if key == "Name Server":
                    if ihunt.data.name_servers is None:
                        ihunt.data.name_servers = [val]
                    else:
                        if val not in ihunt.data.name_servers:
                            ihunt.data.name_servers.append(val)
                if key == "DNSSEC":
                    ihunt.data.dnssec = val
    except Exception as e:
        echo(f"[x] whois command error: {e}", ihunt.verbose)

    echo("[*] Finished executing whois command.", ihunt.verbose)


# Query: IP
# Return: Info
def exec_whois_ip(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Executing whois command...", ihunt.verbose)

    result = subprocess.run(['whois', ihunt.query.value], capture_output=True, text=True)
    lines = result.stdout.splitlines()

    try:
        with lock:
            for line in lines:
                if ':' not in line or '%' in line:
                    continue
                key, val = map(str.strip, line.split(':', 1))
                if key == "inetnum" and ihunt.data.net_range is None:
                    ihunt.data.net_range = val
                if key == "netname" and ihunt.data.net_name is None:
                    ihunt.data.net_name = val
                if key == "descr" and ihunt.data.net_desc is None:
                    ihunt.data.net_desc = val
                if key == "country" and ihunt.data.country_code is None:
                    ihunt.data.country_code = val
                if key == "admin-c":
                    if ihunt.data.net_admin_handle is None:
                        ihunt.data.net_admin_handle = val
                    # Additional information for the admin-c
                    result = subprocess.run(['whois', val], capture_output=True, text=True)
                    lines_2 = result.stdout.splitlines()
                    for line_2 in lines_2:
                        if ':' not in line_2 or '%' in line_2:
                            continue
                        key_2, val_2 = map(str.strip, line_2.split(':', 1))
                        if key_2 == "role" and ihunt.data.net_admin_role is None:
                            ihunt.data.net_admin_role = val_2
                        if key_2 == "person":
                            if ihunt.data.net_admin_persons is None:
                                ihunt.data.net_admin_persons = [val_2]
                            else:
                                ihunt.data.net_admin_persons.append(val_2)
                if key == "tech-c":
                    if ihunt.data.net_tech_handle is None:
                        ihunt.data.net_tech_handle = val
                    # Additional information for the tech-c
                    result = subprocess.run(['whois', val], capture_output=True, text=True)
                    line_2 = result.stdout.splitlines()

                    for line_2 in lines_2:
                        if ':' not in line_2 or '%' in line_2:
                            continue
                        key_2, val_2 = map(str.strip, line_2.split(':', 1))
                        if key_2 == "role" and ihunt.data.net_tech_role is None:
                            ihunt.data.net_tech_role = val_2
                        if key_2 == "person":
                            if ihunt.data.net_tech_persons is None:
                                ihunt.data.net_tech_persons = [val_2]
                            else:
                                ihunt.data.net_tech_persons.append(val_2)
    except Exception as e:
        echo(f"[x] whois command error: {e}", ihunt.verbose)

    echo("[*] Finished executing whois command.", ihunt.verbose)



import threading
from .apis.abuseipdb import req_abuseipdb_ip
from .apis.alienvault import req_alienvalut_domain
from .apis.anubis import req_anubis_domain
from .apis.eva import req_eva_email
from .apis.hackertarget import req_hackertarget_domain
from .apis.hunter import req_hunter_domain, req_hunter_email
from .apis.ipapi import req_ipapi_ip
from .apis.genderize import req_genderize_person
from .apis.robtex import req_robtex_ip
from .apis.subdomaincenter import req_subdomaincenter_domain
from .apis.urldna import req_urldna_url
from .apis.virustotal import req_virustotal_domain, req_virustotal_ip, req_virustotal_url
from .apis.whoisxml import req_whoisxml_domain
from .cmds.dig import exec_dig_domain
from .cmds.whois import exec_whois_domain, exec_whois_ip
from .models import QueryType, Ihunt
from .stdout import echo

lock = threading.Lock()


def gather_domain(ihunt: Ihunt) -> None:
    threads = []

    thread = threading.Thread(target=exec_dig_domain, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    thread = threading.Thread(target=exec_whois_domain, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    thread = threading.Thread(target=req_alienvalut_domain, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    thread = threading.Thread(target=req_anubis_domain, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    thread = threading.Thread(target=req_hackertarget_domain, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    if ihunt.apikeys.hunter:
        thread = threading.Thread(target=req_hunter_domain, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] Hunter API key is not set.", ihunt.verbose)

    thread = threading.Thread(target=req_subdomaincenter_domain, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    if ihunt.apikeys.virustotal:
        thread = threading.Thread(target=req_virustotal_domain, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] VirusTotal API key is not set.", ihunt.verbose)

    if ihunt.apikeys.whoisxml:
        thread = threading.Thread(target=req_whoisxml_domain, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] WhoisXML API key is not set.", ihunt.verbose)

    for thread in threads:
        thread.join()


def gather_email(ihunt: Ihunt) -> None:
    threads = []

    thread = threading.Thread(target=req_eva_email, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    if ihunt.apikeys.hunter:
        thread = threading.Thread(target=req_hunter_email, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] Hunter API key is not set.", ihunt.verbose)

    for thread in threads:
        thread.join()


def gather_ip(ihunt: Ihunt) -> None:
    threads = []

    thread = threading.Thread(target=exec_whois_ip, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    if ihunt.apikeys.abuseipdb:
        thread = threading.Thread(target=req_abuseipdb_ip, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] AbuseIPDB API key is not set.", ihunt.verbose)

    thread = threading.Thread(target=req_ipapi_ip, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    thread = threading.Thread(target=req_robtex_ip, args=(ihunt, lock))
    threads.append(thread)
    thread.start()

    if ihunt.apikeys.virustotal:
        thread = threading.Thread(target=req_virustotal_ip, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] VirusTotal API key is not set.", ihunt.verbose)

    for thread in threads:
        thread.join()


def gather_org(ihunt: Ihunt) -> None:
    threads = []

    for thread in threads:
        thread.join()


def gather_person(ihunt: Ihunt) -> None:
    threads = []

    if ihunt.apikeys.huggingface:
        thread = threading.Thread(target=req_genderize_person, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] Hugging Face API key is not set.", ihunt.verbose)

    for thread in threads:
        thread.join()


def gather_url(ihunt: Ihunt) -> None:
    threads = []

    if ihunt.apikeys.urldna:
        thread = threading.Thread(target=req_urldna_url, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] urlDNA API key is not set.", ihunt.verbose)

    if ihunt.apikeys.virustotal:
        thread = threading.Thread(target=req_virustotal_url, args=(ihunt, lock))
        threads.append(thread)
        thread.start()
    else:
        echo("[x] VirusTotal API key is not set.", ihunt.verbose)

    for thread in threads:
        thread.join()


    # Additional research
    if ihunt.data.dom is not None:
        # Find vulnerabilities of website via HTML (DOM) using LLM
        # ...
        pass


def gather_misc(ihunt: Ihunt) -> None:
    pass


def gather(ihunt: Ihunt) -> None:
    echo("[*] Start gathering information...", ihunt.verbose)

    if ihunt.query.type == QueryType.DOMAIN:
        gather_domain(ihunt)
    elif ihunt.query.type == QueryType.EMAIL:
        gather_email(ihunt)
    elif ihunt.query.type == QueryType.IP:
        gather_ip(ihunt)
    elif ihunt.query.type == QueryType.URL:
        gather_url(ihunt)
    elif ihunt.query.type == QueryType.ORG:
        gather_org(ihunt)
    elif ihunt.query.type == QueryType.PERSON:
        gather_person(ihunt)
    elif ihunt.query.type == QueryType.MISC:
        gather_misc(ihunt)
    elif ihunt.query.type == QueryType.UNKNOWN:
        echo("[x] Query Type Unknown", ihunt.verbose)

    echo("[*] Fished gathering.", ihunt.verbose)



# Docs: https://hackertarget.com/ip-tools/

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://api.hackertarget.com/hostsearch"


# Query: Domain
# Return: Subdomains, IPs
def req_hackertarget_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching HackerTarget...", ihunt.verbose)

    url = BASE_URL + f"/?q={ihunt.query.value}"

    try:
        resp = requests.get(url)
        if resp.status_code == 200:
            with lock:
                lines = resp.content.decode().split('\n')
                for line in lines:
                    spl = line.split(',')
                    # Subdomain
                    if ihunt.data.subdomains is None:
                        ihunt.data.subdomains = [spl[0]]
                    else:
                        if spl[0] not in ihunt.data.subdomains:
                            ihunt.data.subdomains.append(spl[0])
                    # IP
                    if len(spl) == 2:
                        if ihunt.data.ips is None:
                            ihunt.data.ips = [spl[1]]
                        else:
                            if spl[1] not in ihunt.data.ips:
                                ihunt.data.ips.append(spl[1])
    except Exception as e:
        echo(f"[x] HackerTarget API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching HackerTarget.", ihunt.verbose)



# Docs: https://docs.virustotal.com/reference/domain-info

import requests
from threading import Lock
import base64
import urllib.parse
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://www.virustotal.com/api/v3"


# Query: Domain
# Return: Info
def req_virustotal_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching VirusTotal...", ihunt.verbose)

    url = BASE_URL + f"/domains/{ihunt.query.value}"
    headers = {
        "accept": "application/json",
        "x-apikey": ihunt.apikeys.virustotal,
    }

    try:
        resp = requests.get(url, headers=headers)
        with lock:
            if resp.status_code == 200:
                d = resp.json()["data"]
                if ihunt.data.domain is None:
                    ihunt.data.domain = d["id"]
                if ihunt.data.virustotal_link is None:
                    ihunt.data.virustotal_link = d["links"]["self"]
                if ihunt.data.jarm is None:
                    ihunt.data.jarm = d["attributes"]["jarm"]
                if ihunt.data.virustotal_stats is None:
                    ihunt.data.virustotal_stats = d["attributes"]["last_analysis_stats"]
                if ihunt.data.virustotal_analysis is None:
                    ihunt.data.virustotal_analysis = d["attributes"]["last_analysis_results"]
                if ihunt.data.virustotal_votes is None:
                    ihunt.data.virustotal_votes = d["attributes"]["total_votes"]
                if ihunt.data.https_cert_signature_algorithm is None:
                    ihunt.data.https_cert_signature_algorithm = d["attributes"]["last_https_certificate"]["cert_signature"]["signature_algorithm"]
                if ihunt.data.https_cert_signature is None:
                    ihunt.data.https_cert_signature = d["attributes"]["last_https_certificate"]["cert_signature"]["signature"]
    except Exception as e:
        echo(f"[x] VirusTotal API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching VirusTotal.", ihunt.verbose)


# Query: IP
# Return: Info
def req_virustotal_ip(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching VirusTotal...", ihunt.verbose)

    url = BASE_URL + f"/ip_addresses/{ihunt.query.value}"
    headers = {
        "accept": "application/json",
        "x-apikey": ihunt.apikeys.virustotal,
    }

    try:
        resp = requests.get(url, headers=headers)
        with lock:
            if resp.status_code == 200:
                d = resp.json()["data"]
                if ihunt.data.ip is None:
                    ihunt.data.ip = d["id"]
                if ihunt.data.virustotal_link is None:
                    ihunt.data.virustotal_link = d["links"]["self"]
                if ihunt.data.asn is None:
                    ihunt.data.asn = d["attributes"]["asn"]
                if ihunt.data.virustotal_analysis is None:
                    ihunt.data.virustotal_analysis = d["attributes"]["last_analysis_results"]
                if ihunt.data.jarm is None:
                    ihunt.data.jarm = d["attributes"]["jarm"]
                if ihunt.data.country_code is None:
                    ihunt.data.country_code = d["attributes"]["country"]
                if ihunt.data.net_range is None:
                    ihunt.data.net_range = d["attributes"]["network"]
                if ihunt.data.virustotal_stats is None:
                    ihunt.data.virustotal_stats = d["attributes"]["last_analysis_stats"]
                if ihunt.data.virustotal_votes is None:
                    ihunt.data.virustotal_votes = d["attributes"]["total_votes"]
                if ihunt.data.https_cert_signature_algorithm is None:
                    ihunt.data.https_cert_signature_algorithm = d["attributes"]["last_https_certificate"]["cert_signature"]["signature_algorithm"]
                if ihunt.data.https_cert_signature is None:
                    ihunt.data.https_cert_signature = d["attributes"]["last_https_certificate"]["cert_signature"]["signature"]
    except Exception as e:
        echo(f"[x] VirusTotal API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching VirusTotal.", ihunt.verbose)


# Query: URL
# Return: Info
def req_virustotal_url(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching VirusTotal...", ihunt.verbose)

    target_url = urllib.parse.quote(base64.b64encode(ihunt.query.value.encode('utf-8')).decode('utf-8'))

    url = BASE_URL + f"/urls/{target_url}"
    headers = {
        "accept": "application/json",
        "x-apikey": ihunt.apikeys.virustotal,
    }

    try:
        resp = requests.get(url, headers=headers)
        if resp.status_code == 200:
            with lock:
                d = resp.json()["data"]
                if ihunt.data.virustotal_link is None:
                    ihunt.data.virustotal_link = d["links"]["self"]
                if ihunt.data.url is None:
                    ihunt.data.url = d["attributes"]["url"]
                if ihunt.data.virustotal_votes is None:
                    ihunt.data.virustotal_votes = d["attributes"]["total_votes"]
                if ihunt.data.virustotal_stats is None:
                    ihunt.data.virustotal_stats = d["attributes"]["last_analysis_stats"]
                if ihunt.data.virustotal_analysis is None:
                    ihunt.data.virustotal_analysis = d["attributes"]["last_analysis_results"]
                if ihunt.data.virustotal_threat_names is None:
                    ihunt.data.virustotal_threat_names = d["attributes"]["threat_names"]
    except Exception as e:
        echo(f"[x] VirusTotal API error {e}", ihunt.verbose)

    echo("[*] finished fetching VirusTotal.", ihunt.verbose)


from __future__ import annotations
import click
from dataclasses import asdict, dataclass
from enum import Enum
import json
import os
from pprint import PrettyPrinter
from typing import Any
import yaml
from .stdout import echo
from .utils import remove_null_values_in_dict


class QueryType(Enum):
    DOMAIN  = 1
    EMAIL   = 2
    IP      = 3
    URL     = 4
    # Misc types
    ORG     = 5
    PERSON  = 6
    MISC    = 7
    UNKNOWN = 8


@dataclass
class Query:
    value: str
    type: QueryType

    def type_str(self) -> str:
        if self.type == QueryType.DOMAIN:
            return "Domain"
        elif self.type == QueryType.EMAIL:
            return "Email"
        elif self.type == QueryType.IP:
            return "IP"
        elif self.type == QueryType.URL:
            return "URL"
        elif self.type == QueryType.ORG:
            return "Organization"
        elif self.type == QueryType.PERSON:
            return "Person"
        elif self.type == QueryType.MISC:
            return "Misc"
        elif self.type == QueryType.UNKNOWN:
            return "Unknown"
        

class ApiKeyEnvs(Enum):
    ABUSEIPDB = "IHUNT_APIKEY_ABUSEIPDB"
    HUGGINGFACE = "IHUNT_APIKEY_HUGGINGFACE"
    HUNTER = "IHUNT_APIKEY_HUNTER"
    SHODAN = "IHUNT_APIKEY_SHODAN"
    URLDNA = "IHUNT_APIKEY_URLDNA"
    URLSCAN = "IHUNT_APIKEY_URLSCAN"
    VIRUSTOTAL = "IHUNT_APIKEY_VIRUSTOTAL"
    WHOISXML = "IHUNT_APIKEY_WHOISXML"
        

@dataclass
class ApiKeys:
    abuseipdb: str | None = None
    huggingface: str | None = None
    hunter: str | None = None
    shodan: str | None = None
    urldna: str | None = None
    urlscan: str | None = None
    virustotal: str | None = None
    whoisxml: str | None = None


@dataclass
class DataDomain:
    domain: str | None = None
    status: list[str] | None = None
    availability: str | None = None
    ips: list[str] | None = None
    subdomains: list[str] | None = None
    updated_date: str | None = None
    creation_date: str | None = None
    registry_expiry_date: str | None = None
    registrar_name: str | None = None
    registrar_iana_id: str | None = None
    registrant_name: str | None = None
    registrant_organization: str | None = None
    registrant_desc: str | None = None
    registrant_industry: str | None = None
    registrant_country: str | None = None
    registrant_country_code: str | None = None
    registrant_state: str | None = None
    registrant_city: str | None = None
    registrant_street: str | None = None
    registrant_postal_code: str | None = None
    registrant_email: str | None = None
    registrant_fax: str | None = None
    registrant_phone: str | None = None
    registrant_twitter: str | None = None
    registrant_facebook: str | None = None
    registrant_linkedin: str | None = None
    registrant_instagram: str | None = None
    registrant_youtube: str | None = None
    admin_name: str | None = None
    admin_organization: str | None = None
    admin_state: str | None = None
    admin_country: str | None = None
    admin_country_code: str | None = None
    admin_city: str | None = None
    admin_street: str | None = None
    admin_postal_code: str | None = None
    admin_email: str | None = None
    admin_fax: str | None = None
    admin_phone: str | None = None
    tech_name: str | None = None
    tech_organization: str | None = None
    tech_country: str | None = None
    tech_country_code: str | None = None
    tech_state: str | None = None
    tech_city: str | None = None
    tech_street: str | None = None
    tech_postal_code: str | None = None
    tech_email: str | None = None
    tech_fax: str | None = None
    tech_phone: str | None = None
    name_servers: list[str] | None = None
    dnssec: str | None = None
    https_cert_signature_algorithm: str | None = None
    https_cert_signature: str | None = None
    emails: list[str] | None = None
    jarm: str | None = None
    virustotal_link: str | None = None
    virustotal_stats: dict[str, int] | None = None
    virustotal_analysis: dict[str, dict[str, str]] | None = None
    virustotal_votes: dict[str, int] | None = None


@dataclass
class DataEmail:
    email: str | None = None
    domain: str | None = None
    provider_name: str | None = None
    creation_date: str | None = None
    owner_name: str | None = None
    gibberish: bool | None = None
    disposable: bool | None = None
    webmail: bool | None = None
    mx_records: bool | None = None
    smtp_server: bool | None = None
    smtp_check: bool | None = None
    accept_all: bool | None = None
    block: bool | None = None
    spam: bool | None = None


@dataclass
class DataIp:
    ip: str | None = None
    net_range: str | None = None
    net_name: str | None = None
    net_desc: str | None = None
    net_admin_handle: str | None = None
    net_admin_role: str | None = None
    net_admin_persons: list[str] | None = None
    net_tech_handle: str | None = None
    net_tech_role: str | None = None
    net_tech_persons: list[str] | None = None
    asn: Any | None = None
    asname: str | None = None
    organization: str | None = None
    whitelisted: bool | None = None
    country_code: str | None = None
    country_name: str | None = None
    region: str | None = None
    city: str | None = None
    postal_code: str | None = None
    latitude: str | None = None
    longitude: str | None = None
    timezone: str | None = None
    currency: str | None = None
    currency_name: str | None = None
    languages: list[str] | None = None
    usage_type: str | None = None
    isp: str | None = None
    domain: str | None = None
    hostnames: list[str] | None = None
    ports: list[int] | None = None
    is_tor: bool | None = None
    subdomains: list[DataDomain] | None = None
    https_cert_signature_algorithm: str | None = None
    https_cert_signature: str | None = None
    jarm: str | None = None
    virustotal_link: str | None = None
    virustotal_stats: dict[str, int] | None = None
    virustotal_analysis: dict[str, dict[str, str]] | None = None
    virustotal_votes: dict[str, int] | None = None
    

@dataclass
class DataOrg:
    organization: str | None = None
    desc: str | None = None
    founded_year: int | None = None
    address: str | None = None
    country_name: str | None = None
    country_code: str | None = None
    region: str | None = None
    state: str | None = None
    city: str | None = None
    postal_code: str | None = None
    latitude: str | None = None
    longitude: str | None = None
    emails: list[str] | None = None
    faxes: list[str] | None = None
    phones: list[str] | None = None
    domains: list[str] | None = None
    websites: list[str] | None = None
    ceo: str | None = None
    employees: list[str] | None = None
    services: list[str] | None = None
    parent_organization: str | None = None
    

@dataclass
class DataPerson:
    person: str | None = None
    aliases: list[str] | None = None
    age: int | None = None
    gender: str | None = None
    nationality: str | None = None
    birth_date: str | None = None
    birth_place: str | None = None
    address: str | None = None
    prev_addresses: list[str] | None = None
    education: str | None = None
    occupation: str | None = None
    employment_history: str | None = None
    acquaintances: list[str] | None = None
    friends: list[str] | None = None
    family_members: str | None = None
    emails: list[str] | None = None
    faxes: list[str] | None = None
    phones: list[str] | None = None
    websites: list[str] | None = None
    social_accounts: dict[str, str] | None = None
    photos: list[Any] | None = None
    hobbies: list[str] | None = None
    languages: list[str] | None = None
    race: str | None = None
    hair_color: str | None = None
    height: str | None = None
    weight: str | None = None
    personality: str | None = None


@dataclass
class DataUrl:
    url: str | None = None
    cert_issuer: str | None = None
    cert_subject: str | None = None
    cert_serial_number: str | None = None
    console_messages: list[dict[str, str]] | None = None
    cookies: list[dict[str, Any]] | None = None
    dom: str | None = None
    favicon: Any | None = None
    ip: str | None = None
    asn: Any | None = None
    isp: str | None = None
    organization: str | None = None
    country_name: str | None = None
    country_code: str | None = None
    region: str | None = None
    city:  str | None = None
    latitude: str | None = None
    longitude: str | None = None
    malicious: Any | None = None
    page: Any | None = None
    screenshot: Any | None = None
    technologies: list[Any] | None = None
    virustotal_link: str | None = None
    virustotal_stats: dict[str, int] | None = None
    virustotal_analysis: dict[str, dict[str, str]] | None = None
    virustotal_votes: dict[str, int] | None = None
    virustotal_threat_names: list[str] | None = None


@dataclass
class DataMisc:
    entity: str | None = None


def init_data(query: Query) -> DataDomain | DataEmail | DataIp | DataOrg | DataPerson | DataUrl:
    if query.type == QueryType.DOMAIN:
        return DataDomain(domain=query.value)
    elif query.type == QueryType.EMAIL:
        return DataEmail(email=query.value)
    elif query.type == QueryType.IP:
        return DataIp(ip=query.value)
    elif query.type == QueryType.ORG:
        return DataOrg(organization=query.value)
    elif query.type == QueryType.PERSON:
        return DataPerson(person=query.value)
    elif query.type == QueryType.URL:
        return DataUrl(url=query.value)
    elif query.type == QueryType.MISC:
        return DataMisc(entity=query.value)


class Ihunt:
    def __init__(self, query: Query, depth: int, format: str, output: str, verbose: bool) -> None:
        self.query: Query = query
        self.depth: int = depth
        self.format: str = format
        self.output: str = output
        self.verbose: bool = verbose
        self.apikeys = ApiKeys(
            abuseipdb=os.getenv(ApiKeyEnvs.ABUSEIPDB.value),
            huggingface=os.getenv(ApiKeyEnvs.HUGGINGFACE.value),
            hunter=os.getenv(ApiKeyEnvs.HUNTER.value),
            shodan=os.getenv(ApiKeyEnvs.SHODAN.value),
            urldna=os.getenv(ApiKeyEnvs.URLDNA.value),
            urlscan=os.getenv(ApiKeyEnvs.URLSCAN.value),
            virustotal=os.getenv(ApiKeyEnvs.VIRUSTOTAL.value),
            whoisxml=os.getenv(ApiKeyEnvs.WHOISXML.value),
        )
        self.data = init_data(query)

    def print_options(self) -> None:
        options = f"""
+ Query         : {self.query.value}
+ Query Type    : {self.query.type_str()}
+ Depth         : {self.depth}
+ Format        : {self.format}
+ Output        : {self.output}
+ Verbose       : {self.verbose}
"""
        echo(options, self.verbose)
    
    def print_data(self) -> None:
        echo("[*] Results:\n", self.verbose)

        data = remove_null_values_in_dict(asdict(self.data))

        if self.format == "json":
            click.echo(json.dumps(data, indent=4))
        elif self.format == "pretty":
            click.echo(PrettyPrinter(indent=4).pformat(data))
        elif self.format == "yaml":
            click.echo(yaml.dump(data, default_flow_style=False, indent=4))
        
    def write(self) -> None:
        echo(f"[*] Writing results to {self.output}...", self.verbose)

        data = remove_null_values_in_dict(asdict(self.data))

        with open(self.output, 'w') as f:
            output_ext = os.path.splitext(self.output)
            if output_ext[1] == ".json":
                f.write(json.dumps(data, indent=4))
            elif output_ext[1] == ".yaml" or output_ext[1] == ".yml":
                f.write(yaml.dump(data, default_flow_style=False, indent=4))
            else:
                f.write(PrettyPrinter(indent=4).pformat(data))




# Docs: https://www.robtex.com/api/

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://freeapi.robtex.com/ipquery"


# Query: IP
# Return: Info
def req_robtex_ip(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching Robtex...", ihunt.verbose)

    url = BASE_URL + f"/{ihunt.query.value}"

    try:
        resp = requests.get(url)
        if resp.status_code == 200:
            with lock:
                d = resp.json()
                if d["status"] != "ok":
                    return
                if ihunt.data.city is None:
                    ihunt.data.city = d["city"]
                if ihunt.data.country_name is None:
                    ihunt.data.country_name = d["country"]
                if ihunt.data.asn is None:
                    ihunt.data.asn = d["as"]
                if ihunt.data.asname is None:
                    ihunt.data.asname = d["asname"]
    except Exception as e:
        echo(f"[x] Robtex API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching Robtex...", ihunt.verbose)
        


# Docs: https://otx.alienvault.com/api

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo
from ..utils import is_ip_address

BASE_URL = "https://otx.alienvault.com/api/v1/indicators/domain"


# Query: Domain
# Return: Subdomains
def req_alienvalut_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching AlienVault...", ihunt.verbose)

    url = BASE_URL + f"/{ihunt.query.value}/passive_dns"

    try:
        resp = requests.get(url)
        if resp.status_code == 200:
            with lock:
                data = resp.json()["passive_dns"]
                for d in data:
                    if is_ip_address(d["address"]):
                        if ihunt.data.ips is None:
                            ihunt.data.ips = [d["address"]]
                        else:
                            if d["address"] not in ihunt.data.ips:
                                ihunt.data.ips.append(d["address"])
                    if is_ip_address(d["hostname"]) is False:
                        if ihunt.data.subdomains is None:
                            ihunt.data.subdomains = [d["hostname"]]
                        else:
                            if d["hostname"] not in ihunt.data.subdomains:
                                ihunt.data.subdomains.append(d["hostname"])
    except Exception as e:
        echo(f"[x] AlienVault API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching AlienVault.", ihunt.verbose)
        


import click
from dotenv import load_dotenv
import os
import re
from .gather import gather
from .models import Query, QueryType, Ihunt
from .querytype import identify_querytype


def print_version(ctx, param, value) -> None:
    if not value or ctx.resilient_parsing:
        return
    version = "0.0.0"
    click.echo(f"ihunt v{version}")
    ctx.exit()


@click.command()
@click.option('-c', '--config', help='Config file.')
@click.option('-d', '--depth', type=click.IntRange(1, 3), default=1, help='Depth of information gathering.')
@click.option('-f', '--format', type=click.Choice(['json', 'pretty', 'yaml']), default='json', help='Print format.')
@click.option('-o', '--output', help='Write results to the output file path.')
@click.option('-v', '--verbose', is_flag=True, help='Verbose output.')
@click.option('--version', is_flag=True, callback=print_version, expose_value=False, is_eager=True, help='The version of the Ihunt.')
@click.argument('query')
def run(config: str, depth: int, format: str, output: str, verbose: bool, query: str) -> None:
    if config is not None:
        # Load configurations from file.
        if os.path.exists(config) is False:
            click.echo(f"[x] The config file does not exist: {config}")
            return
        load_dotenv(config)

    query = re.sub(r'\s+', ' ', query).strip()

    ihunt = Ihunt(
        query=Query(value=query, type=identify_querytype(query, verbose)),
        depth=depth,
        format=format,
        output=output,
        verbose=verbose,
    )
    ihunt.print_options()

    if ihunt.query.type == QueryType.UNKNOWN:
        click.echo(f"Unknown query type: {ihunt.query.value}")
        return

    gather(ihunt)

    if ihunt.output is None:
        ihunt.print_data()
    else:
        ihunt.write()


def main() -> int:
    run()
    return 0



import ipaddress
from typing import Any


def is_ip_address(s: str) -> bool:
    try:
        ipaddress.ip_address(s)
        return True
    except:
        return False


def has_duplicate(arr: list[Any]) -> bool:
    return len(arr) != len(set(arr))


# The function is used when stdout or write to the output file.
def remove_null_values_in_dict(data: dict[Any]) -> dict[Any]:
    return {
        k: remove_null_values_in_dict(v)
        if isinstance(v, dict)
        else v
        for k, v in data.items()
        if v is not None
    }


# Docs: https://whois.whoisxmlapi.com/documentation/making-requests

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://www.whoisxmlapi.com/whoisserver/WhoisService"


# Query: Domain
# Return: Info
def req_whoisxml_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching WhoisXML...", ihunt.verbose)

    url = BASE_URL
    headers = {
        "Content-Type": "application/json",
    }
    data = {
        "domainName": ihunt.query.value,
        "outputFormat": "JSON",
        "apiKey": ihunt.apikeys.whoisxml,
    }

    try:
        resp = requests.post(url, headers=headers, json=data)
        if resp.status_code == 200:
            with lock:
                d = resp.json()["WhoisRecord"]
                if ihunt.data.domain is None:
                    ihunt.data.domain = d["domainName"]
                if ihunt.data.status is None:
                    ihunt.data.status = [d["status"]]
                # if ihunt.data.availability is None:
                #     ihunt.data.availability = d["domainAvailability"]
                # if ihunt.data.ips is None:
                #     ihunt.data.domain = d["ips"]
                if ihunt.data.registrar_name is None:
                    ihunt.data.registrar_name = d["registryData"]["registrarName"]
                if ihunt.data.registrar_iana_id is None:
                    ihunt.data.registrar_iana_id = d["registryData"]["registrarIANAID"]
                if ihunt.data.registrant_organization is None:
                    ihunt.data.registrant_organization = d["registrant"]["organization"]
                if ihunt.data.registrant_country is None:
                    ihunt.data.registrant_country = d["registrant"]["country"]
                if ihunt.data.registrant_country_code is None:
                    ihunt.data.registrant_country_code = d["registrant"]["countryCode"]
                if ihunt.data.registrant_state is None:
                    ihunt.data.registrant_state = d["registrant"]["state"]
                if ihunt.data.registrant_city is None:
                    ihunt.data.registrant_city = d["registrant"]["city"]
                if ihunt.data.admin_organization is None:
                    ihunt.data.admin_organization = d["administrativeContact"]["organization"]
                if ihunt.data.admin_state is None:
                    ihunt.data.admin_state = d["administrativeContact"]["state"]
                if ihunt.data.admin_country is None:
                    ihunt.data.admin_country = d["administrativeContact"]["country"]
                if ihunt.data.admin_country_code is None:
                    ihunt.data.admin_country_code = d["administrativeContact"]["countryCode"]
                if ihunt.data.tech_organization is None:
                    ihunt.data.tech_organization = d["technicalContact"]["organization"]
                if ihunt.data.tech_country is None:
                    ihunt.data.tech_country = d["technicalContact"]["country"]
                if ihunt.data.tech_country_code is None:
                    ihunt.data.tech_country_code = d["technicalContact"]["countryCode"]
                if ihunt.data.tech_state is None:
                    ihunt.data.tech_state = d["technicalContact"]["state"]
                if ihunt.data.name_servers is None:
                    ihunt.data.name_servers = d["nameServers"]["hostNames"]
    except Exception as e:
        echo(f"[x] WhoisXML API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching WhoisXML...", ihunt.verbose)



import re
from .models import QueryType
from .stdout import echo
from .utils import is_ip_address


# Search models: https://huggingface.co/models?pipeline_tag=token-classification
model_ner = "dslim/bert-base-NER"
# model_ner = "dbmdz/bert-large-cased-finetuned-conll03-english"
    

# Uses LLM NER task.
def identify_entity(query: str) -> QueryType:
    import torch
    from transformers import pipeline
    device = (
        "cuda"
        if torch.cuda.is_available()
        else "mps"
        if torch.backends.mps.is_available()
        else "cpu"
    )
    ner = pipeline(task="ner", model=model_ner, device=device)
    entities = ner(query)
    
    if len(entities) == 0:
        return QueryType.UNKNOWN

    for entity in entities:
        entity_type = entity['entity']
        if "-ORG" in entity_type:
            return QueryType.ORG
        elif "-PER" in entity_type:
            return QueryType.PERSON
        else:
            return QueryType.MISC


def identify_querytype(query: str, verbose: bool) -> QueryType:
    echo("[*] Identifying the query type...", verbose)

    url_pattern = re.compile(r'^(https?|ftp|ssh)://[^\s/$.?#].[^\s]*$', re.IGNORECASE)
    email_pattern = re.compile(r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$')
    domain_pattern = re.compile(r'^((?!-)[A-Za-z0-9-]{1,63}(?<!-)\.)?(?!-)[A-Za-z0-9-]{1,63}(?<!-)\.(?!-)[A-Za-z0-9-]{1,63}(?<!-)$')

    if domain_pattern.match(query):
        return QueryType.DOMAIN
    elif email_pattern.match(query):
        return QueryType.EMAIL
    elif is_ip_address(query):
        return QueryType.IP
    elif url_pattern.match(query):
        return QueryType.URL
    else:
        # Identify it using LLM
        return identify_entity(query)



import re
import subprocess
from threading import Lock
from ..models import Ihunt
from ..stdout import echo


# Query: Domain
# Return: Subdomains
def exec_dig_domain(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Executing dig command...", ihunt.verbose)

    result = subprocess.run(['dig', '+short', ihunt.query.value], capture_output=True, text=True)
    lines = result.stdout.splitlines()

    try:
        with lock:
            for line in lines:
                if re.match(r'^\d{1,3}(\.\d{1,3}){3}$', line):
                    if ihunt.data.ips is None:
                        ihunt.data.ips = [line]
                    else:
                        if line not in ihunt.data.ips:
                            ihunt.data.ips.append(line)
    except Exception as e:
        echo(f"[x] dig command error: {e}", ihunt.verbose)

    echo("[*] Finished executing dig command.", ihunt.verbose)



# Reference: https://docs.abuseipdb.com/#introduction

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://api.abuseipdb.com/api/v2"


# Query: IP
# Return: Info
def req_abuseipdb_ip(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching AbuseIPDB...", ihunt.verbose)

    url = BASE_URL + "/check"

    headers = {
        "Key": ihunt.apikeys.abuseipdb,
        "Accept": "application/json",
    }
    params = {
        "ipAddress": ihunt.query.value,
        "maxAgeInDays": 90,
        "verbose": "",
    }

    try:
        resp = requests.get(url, headers=headers, params=params)
        if resp.status_code == 200:
            with lock:
                d = resp.json()["data"]
                if ihunt.data.ip is None:
                    ihunt.data.ip = d["ipAddress"]
                if ihunt.data.whitelisted is None:
                    ihunt.data.whitelisted = d["isWhitelisted"]
                if ihunt.data.country_code is None:
                    ihunt.data.country_code = d["countryCode"]
                if ihunt.data.country_name is None:
                    ihunt.data.country_name = d["countryName"]
                if ihunt.data.usage_type is None:
                    ihunt.data.usage_type = d["usageType"]
                if ihunt.data.isp is None:
                    ihunt.data.isp = d["isp"]
                if ihunt.data.domain is None:
                    ihunt.data.domain = d["domain"]
                if ihunt.data.hostnames is None:
                    ihunt.data.hostnames = d["hostnames"]
                if ihunt.data.is_tor is None:
                    ihunt.data.is_tor = d["isTor"]
    except Exception as e:
        echo(f"[x] AbuseIPDB API error: {e}", ihunt.verbose)

    echo("[*] Fishined fetching AbuseIPDB.", ihunt.verbose)



# Docs: https://ipapi.co/?q=78.48.50.249

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://ipapi.co"


# Query: IP
# Return: Info
def req_ipapi_ip(ihunt: Ihunt, lock: Lock) -> None:
    echo("[*] Fetching IPAPI...", ihunt.verbose)

    url = BASE_URL + f"/{ihunt.query.value}/json/"

    try:
        resp = requests.get(url)
        if resp.status_code == 200:
            with lock:
                d = resp.json()
                if ihunt.data.ip is None:
                    ihunt.data.ip = d["ip"]
                if ihunt.data.net_range is None:
                    ihunt.data.net_range = d["network"]
                if ihunt.data.region is None:
                    ihunt.data.region = d["region"]
                if ihunt.data.country_code is None:
                    ihunt.data.country_code = d["country_code"]
                if ihunt.data.country_name is None:
                    ihunt.data.country_name = d["country_name"]
                if ihunt.data.postal_code is None:
                    ihunt.data.postal_code = d["postal"]
                if ihunt.data.latitude is None:
                    ihunt.data.latitude = d["latitude"]
                if ihunt.data.longitude is None:
                    ihunt.data.longitude = d["longitude"]
                if ihunt.data.timezone is None:
                    ihunt.data.timezone = d["timezone"]
                if ihunt.data.currency is None:
                    ihunt.data.currency = d["currency"]
                if ihunt.data.currency_name is None:
                    ihunt.data.currency_name = d["currency_name"]
                if ihunt.data.languages is None:
                    ihunt.data.languages = d["languages"].split(',')
                else:
                    for lang in d["languages"].split(','):
                        if lang not in ihunt.data.languages:
                            ihunt.data.languages.append(lang)
                if ihunt.data.asn is None:
                    ihunt.data.asn = d["asn"]
                if ihunt.data.organization is None:
                    ihunt.data.organization = d["org"]
    except Exception as e:
        echo(f"[x] IPAPI API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching IPAPI.", ihunt.verbose)



# from huggingface_hub import login
# from threading import Lock
# from transformers import pipeline
# from ..models import Ihunt
# from ..stdout import echo

# # Query: Organization
# # Return: The organization7s summary
# def req_huggingface_org(ihunt: Ihunt, lock: Lock) -> None:
#     echo("[*] Fetching Hugging Face...", ihunt.verbose)

#     login(token=ihunt.apikeys.huggingface)

#     pipe = pipeline(
#         "text-generation", model="distilbert/distilgpt2")

#     content = f"""
# Please tell me information about {ihunt.query.value}. The format is below:

# - name
# - description
# - founded_year
# - address
# - country_name
# - country_code
# - region
# - state
# - city
# - postal_code
# - latitude
# - longitude
# - emails
# - faxes
# - phones
# - domains
# - websites
# - ceo
# """

#     outputs = pipe(content, max_new_tokens=256)
#     print(outputs)

#     with lock:
#         pass

#     echo("[*] Finished fetching Hugging Face.", ihunt.verbose)


# Docs: https://genderize.io/documentation

import requests
from threading import Lock
from ..models import Ihunt
from ..stdout import echo

BASE_URL = "https://api.genderize.io"


# Query: Person
# Return: Gender
def req_genderize_person(ihunt: Ihunt, lock: Lock) -> None:
    url = BASE_URL
    params = {
        "name": ihunt.query.value,
    }

    try:
        resp = requests.get(url, params=params)
        if resp.status_code == 200:
            with lock:
                d = resp.json()
                if ihunt.data.gender is None:
                    ihunt.data.gender = d["gender"]
    except Exception as e:
        echo(f"[x] Genderize API error: {e}", ihunt.verbose)

    echo("[*] Finished fetching Genderize.", ihunt.verbose)
